@@ -22,23 +22,13 @@
  */
 
 #include <stdio.h>
+#include <ctype.h>
 #include <string.h>
 #include <R.h>
 #include <Rinternals.h>
-//#include "foreign.h"
+#include "foreign.h"
 #include "SASxport.h"
 
-
-#ifdef ENABLE_NLS
-#include <libintl.h>
-#define _(String) dgettext ("foreign", String)
-#define gettext_noop(String) (String)
-#else
-#define _(String) (String)
-#define gettext_noop(String) (String)
-#endif
-
-
 #define HEADER_BEG "HEADER RECORD*******"
 #define HEADER_TYPE_LIBRARY "LIBRARY "
 #define HEADER_TYPE_MEMBER "MEMBER  "
@@ -153,9 +143,9 @@ get_lib_header(FILE *fp, struct SAS_XPORT_header *head)
     int n;
 
     n = GET_RECORD(record, fp, 80);
-    if(n == 80 && strncmp(LIB_HEADER, record, 80) != 0)
+    if(n != 80 || strncmp(LIB_HEADER, record, 80) != 0)
 	error(_("file not in SAS transfer format"));
-  
+
     n = GET_RECORD(record, fp, 80);
     if(n != 80)
 	return 0;
@@ -205,9 +195,12 @@ get_mem_header(FILE *fp, struct SAS_XPORT_member *member)
     n = GET_RECORD(record, fp, 80);
     if(n != 80)
 	return 0;
+    record[80] = '\0';
     memcpy(member->sas_mod, record, 16);
-    if((strrchr(record+16, ' ') - record) != 79)
-	return 0;
+
+    memcpy(member->sas_dslabel, record+32, 40);
+    memcpy(member->sas_dstype, record+72, 8);
+
     return 1;
 }
 
@@ -230,8 +223,9 @@ init_xport_info(FILE *fp)
     Free(lib_head);
 
     n = GET_RECORD(record, fp, 80);
-    if(n != 80 || strncmp(MEM_HEADER, record, 75) != 0 ||
-       strncmp("  ", record+78, 2) != 0)
+    if(n != 80 || 
+       strncmp(MEM_HEADER, record, 75) != 0 ||
+       strncmp("  ", record+78, 2) != 0 )
 	error(_("file not in SAS transfer format"));
     record[78] = '\0';
     sscanf(record+75, "%d", &namestr_length);
@@ -240,7 +234,7 @@ init_xport_info(FILE *fp)
 }
 
 static int
-init_mem_info(FILE *fp, char *name)
+init_mem_info(FILE *fp, char *name, char *dslabel, char *dstype)
 {
     int length, n;
     char record[81];
@@ -263,6 +257,7 @@ init_mem_info(FILE *fp, char *name)
     record[58] = '\0';
     sscanf(record+54, "%d", &length);
 
+    /* Extract data set name */
     tmp = strchr(mem_head->sas_dsname, ' ');
     n = tmp - mem_head->sas_dsname;
     if(n > 0) {
@@ -272,6 +267,22 @@ init_mem_info(FILE *fp, char *name)
 	name[n] = '\0';
     } else name[0] = '\0';
 
+    /* Extract data set label, and trim trailing blanks */
+    strncpy(dslabel, mem_head->sas_dslabel, 40);
+    for(int i=40-1; i>0; i--)
+      if( isspace(dslabel[i]) )
+	dslabel[i] = '\0';
+      else
+	break;
+
+    /* Extract data set type */
+    strncpy(dstype, mem_head->sas_dstype, 8);
+    for(int i=8-1; i>0; i--)
+      if( isspace(dstype[i]) )
+	dstype[i] = '\0';
+      else
+	break;
+
     Free(mem_head);
 
     return length;
@@ -297,7 +308,7 @@ next_xport_info(FILE *fp, int namestr_length, int nvars,
 {
     char *tmp;
     char record[81];
-    int i, n, nbytes, totwidth, nlength, restOfCard;
+    int i, n, totwidth, nlength, restOfCard;
     struct SAS_XPORT_namestr *nam_head;
 
     nam_head = Calloc(nvars, struct SAS_XPORT_namestr);
@@ -394,7 +405,6 @@ next_xport_info(FILE *fp, int namestr_length, int nvars,
     for(i = 0; i < nvars; i++)
 	totwidth += nlng[i];
 
-    nbytes = 0;
     nlength = 0;
     tmp = Calloc(totwidth <= 80 ? 81 : (totwidth+1), char);
     restOfCard = 0;
@@ -433,6 +443,18 @@ next_xport_info(FILE *fp, int namestr_length, int nvars,
 		break;
 	    }
 	}
+	else /* beware that the previous member can end on card
+	      * boundary with no padding */
+	  if (restOfCard == 80 && n == 80 &&
+	      strncmp(MEM_HEADER, tmp, 75) == 0 &&
+	      strncmp("  ", tmp+78, 2) == 0) {
+	    strncpy(record, tmp, 80);
+	    *tailpad = 0;
+	    record[78] = '\0';
+	    sscanf(record+75, "%d", &namestr_length);
+	    break;
+	  }
+	
 	if (fsetpos(fp, &currentPos)) {
 	    error(_("problem accessing SAS XPORT file"));
 	}
@@ -541,8 +563,11 @@ xport_info(SEXP xportFile)
     FILE *fp;
     int i, namestrLength, memLength, ansLength;
     char dsname[9];
+    char dslabel[41];
+    char dstype[9];
     SEXP ans, ansNames, varInfoNames, varInfo;
     SEXP char_numeric, char_character;
+    SEXP dfLabel, dfType;
 
     PROTECT(varInfoNames = allocVector(STRSXP, VAR_INFO_LENGTH));
     for(i = 0; i < VAR_INFO_LENGTH; i++)
@@ -555,18 +580,28 @@ xport_info(SEXP xportFile)
 	error(_("first argument must be a file name"));
     fp = fopen(R_ExpandFileName(CHAR(STRING_ELT(xportFile, 0))), "rb");
     if (!fp)
-	error(_("unable to open file"));
+        error(_("unable to open file: '%s'"), strerror(errno));
     namestrLength = init_xport_info(fp);
 
     ansLength = 0;
     PROTECT(ans = allocVector(VECSXP, 0));
-    PROTECT(ansNames = allocVector(STRSXP, 0));
+    PROTECT(ansNames  = allocVector(STRSXP, 0));
 
-    while(namestrLength > 0 && (memLength = init_mem_info(fp, dsname)) > 0) {
+    while(namestrLength > 0 && (memLength = init_mem_info(fp, dsname, dslabel, dstype)) > 0) {
 
 	PROTECT(varInfo = allocVector(VECSXP, VAR_INFO_LENGTH));
 	setAttrib(varInfo, R_NamesSymbol, varInfoNames);
 
+	dslabel[40] = '\n';
+	PROTECT(dfLabel = allocVector(STRSXP, 1));
+	SET_STRING_ELT(dfLabel, 0, mkChar(dslabel));
+	setAttrib(varInfo, install("label"  ), dfLabel);
+
+	dstype[8] = '\n';
+	PROTECT(dfType  = allocVector(STRSXP, 1));
+	SET_STRING_ELT(dfType, 0, mkChar(dstype));
+	setAttrib(varInfo, install("SAStype"), dfType );
+
 	SET_XPORT_VAR_TYPE(varInfo, allocVector(STRSXP, memLength));
 	SET_XPORT_VAR_WIDTH(varInfo, allocVector(INTSXP, memLength));
 	SET_XPORT_VAR_INDEX(varInfo, allocVector(INTSXP, memLength));
@@ -610,7 +645,8 @@ xport_info(SEXP xportFile)
 			   char_character);
 	}
 	PROTECT(ans = lengthgets(ans, ansLength+1));
-	PROTECT(ansNames = lengthgets(ansNames, ansLength+1));
+	PROTECT(ansNames  = lengthgets(ansNames,  ansLength+1));
+
 /*  	PROTECT(newAns = allocVector(VECSXP, ansLength+1)); */
 /*  	PROTECT(newAnsNames = allocVector(STRSXP, ansLength+1)); */
 
@@ -621,15 +657,16 @@ xport_info(SEXP xportFile)
 /*  	ans = newAns; */
 /*  	ansNames = newAnsNames; */
 
-	SET_STRING_ELT(ansNames, ansLength, mkChar(dsname));
+	SET_STRING_ELT(ansNames , ansLength, mkChar(dsname ));
 	SET_VECTOR_ELT(ans, ansLength, varInfo);
 	ansLength++;
  
-	UNPROTECT(5);
+	UNPROTECT(7);
 	PROTECT(ans);
 	PROTECT(ansNames);
     }
 
+
     setAttrib(ans, R_NamesSymbol, ansNames);  
     UNPROTECT(5);
     fclose(fp);
@@ -659,7 +696,7 @@ xport_read(SEXP xportFile, SEXP xportInfo)
 	error(_("first argument must be a file name"));
     fp = fopen(R_ExpandFileName(CHAR(STRING_ELT(xportFile, 0))), "rb");
     if (!fp)
-	error(_("unable to open file"));
+        error(_("unable to open file: '%s'"), strerror(errno));
     if (fseek(fp, 240, SEEK_SET) != 0)
 	error(_("problem reading SAS XPORT file '%s'"),
 	      CHAR(STRING_ELT(xportFile, 0)));
