@@ -2,9 +2,9 @@
  *
  *    writeSAS.h: Routines for writing SAS XPT formatted files
  *
- *    Author:  Gregory R. Warnes <greg@random-technologies-llc.com>
+ *    Author:  Gregory R. Warnes <greg@warnes.net>
  *
- *    Copyright (C) 2007  Random Technologies LLC
+ *    Copyright (C) 2007  Gregory R. Warnes <greg@warnes.net>
  * 
  *    This program is free software; you can redistribute it and/or modify
  *    it under the terms of the GNU General Public License as published by
@@ -103,7 +103,10 @@ struct MEMBER_HEADER {
 
   /* Line 4 */
   char sas_modified[16];
-  char blanks2[64];   
+  //char blanks2[64];   
+  char padding[16];
+  char dslabel[40];
+  char dstype[8];
 };
 
 
@@ -155,7 +158,7 @@ void zeroCopy(char *target, int len, char *source);
 
 void fill_file_header(char **cDate, char **mDate, char **sasVer, char **osType);
 void fill_member_header(char **dfName, char **sasVer, char **osType, char **cDate, 
-			char **mDate);
+			char **mDate, char **dfLabel, char **dfType);
 
 void fill_namestr(int  *isChar, int  *nlng, int  *nvar0, char **nname, char **nlabel, 
 		  char **nform, int  *nfl, int  *nfd, int  *nfj, char **niform, 
